import uuid
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Sequence
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Generic, cast

from cognite.client.data_classes import filters
from cognite.client.data_classes.aggregations import UniqueResult
from cognite.client.data_classes.assets import AssetProperty
from cognite.client.data_classes.documents import SourceFileProperty
from cognite.client.data_classes.events import EventProperty
from pydantic import JsonValue

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client._resource_base import T_Identifier, T_RequestResource, T_ResponseResource
from cognite_toolkit._cdf_tk.client.resource_classes.apm_config_v1 import (
    APM_CONFIG_SPACE,
    APMConfigResponse,
    Discipline,
    FeatureConfiguration,
    RootLocationConfiguration,
    RootLocationFeatureToggles,
)
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import (
    InstanceSource,
    NodeReference,
    NodeRequest,
    SpaceRequest,
    ViewReference,
)
from cognite_toolkit._cdf_tk.client.resource_classes.dataset import DataSetResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.infield import (
    INFIELD_ON_CDM_DATA_MODEL,
    InFieldCDMLocationConfigRequest,
)
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import TypedNodeIdentifier
from cognite_toolkit._cdf_tk.client.resource_classes.location_filter import LocationFilterRequest
from cognite_toolkit._cdf_tk.cruds import (
    InFieldCDMLocationConfigCRUD,
    LocationFilterCRUD,
    NodeCRUD,
    ResourceCRUD,
    SpaceCRUD,
)
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitMigrationError,
    ToolkitMissingResourceError,
    ToolkitRequiredValueError,
)
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning, LowSeverityWarning
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.text import fix_invalid_space_name, warn_invalid_space_name

from .data_model import CREATED_SOURCE_SYSTEM_VIEW_ID, SPACE, SPACE_SOURCE_VIEW_ID


@dataclass
class CreatedResource(Generic[T_RequestResource]):
    resource: T_RequestResource
    config_data: dict[str, Any] | None = None
    filestem: str | None = None


@dataclass
class ToCreateResources(Generic[T_Identifier, T_RequestResource, T_ResponseResource]):
    resources: Sequence[CreatedResource[T_RequestResource]]
    crud_cls: type[ResourceCRUD[T_Identifier, T_RequestResource, T_ResponseResource]]
    display_name: str
    store_linage: Callable[[], int] | None = None


class MigrationCreator(ABC):
    """Base class for migration resources configurations that are created resources."""

    def __init__(self, client: ToolkitClient) -> None:
        self.client = client

    @abstractmethod
    def create_resources(self) -> Iterable[ToCreateResources]:
        raise NotImplementedError("Subclasses should implement this method")


class InstanceSpaceCreator(MigrationCreator):
    """Creates instance spaces for migration."""

    def __init__(
        self,
        client: ToolkitClient,
        datasets: list[DataSetResponse] | None = None,
        data_set_external_ids: list[str] | None = None,
        auto_fix: bool = False,
    ) -> None:
        super().__init__(client)
        if sum([datasets is not None, data_set_external_ids is not None]) != 1:
            raise ValueError("Exactly one of datasets or data_set_external_ids must be provided.")
        self.data_set_external_ids = data_set_external_ids
        self.datasets = datasets or []
        self.auto_fix = auto_fix

    def create_resources(self) -> Iterable[ToCreateResources]:
        if self.data_set_external_ids is not None:
            self.datasets = self.client.tool.datasets.retrieve(ExternalId.from_external_ids(self.data_set_external_ids))

        if missing_external_ids := [ds.id for ds in self.datasets if ds.external_id is None]:
            raise ToolkitRequiredValueError(
                f"Cannot create instance spaces for datasets with missing external IDs: {humanize_collection(missing_external_ids)}"
            )

        space_id_by_dataset = self._resolve_space_ids()

        created_resources: list[CreatedResource[SpaceRequest]] = []
        linage_nodes: list[NodeRequest] = []
        for dataset in self.datasets:
            data_set_external_id = cast(str, dataset.external_id)
            space_id = space_id_by_dataset[data_set_external_id]
            space = SpaceRequest(
                space=space_id,
                name=dataset.name,
                description=dataset.description,
            )
            linage_node = NodeRequest(
                space=SPACE.space,
                external_id=space.space,
                sources=[
                    InstanceSource(
                        source=SPACE_SOURCE_VIEW_ID,
                        properties={
                            "instanceSpace": space.space,
                            "dataSetId": dataset.id,
                            "dataSetExternalId": data_set_external_id,
                        },
                    )
                ],
            )
            linage_nodes.append(linage_node)
            created_resources.append(
                CreatedResource(
                    resource=space,
                    config_data=space.dump(),
                    filestem=space.space,
                )
            )
        yield ToCreateResources(
            resources=created_resources,
            crud_cls=SpaceCRUD,
            display_name="Instance Space",
            store_linage=lambda: self.store_lineage(linage_nodes),
        )

    def _resolve_space_ids(self) -> dict[str, str]:
        """Compute the space ID for each dataset external ID, applying auto-fix if enabled.

        Raises ToolkitMigrationError if multiple dataset external IDs resolve to the same space ID.
        """
        space_id_to_external_ids: dict[str, list[str]] = {}
        result: dict[str, str] = {}

        for dataset in self.datasets:
            data_set_external_id = cast(str, dataset.external_id)
            warning_message = warn_invalid_space_name(data_set_external_id)
            space_id = data_set_external_id
            if warning_message and self.auto_fix:
                space_id = fix_invalid_space_name(data_set_external_id)
                LowSeverityWarning(
                    f"{warning_message}\nAutomatically fixing space name to {space_id!r}."
                ).print_warning(console=self.client.console)
            elif warning_message:
                HighSeverityWarning(
                    f"{warning_message}\nRun command with `--auto-fix` to automatically make the data set external ID a valid space ID."
                ).print_warning(console=self.client.console)

            result[data_set_external_id] = space_id
            space_id_to_external_ids.setdefault(space_id, []).append(data_set_external_id)

        duplicates = {
            space_id: external_ids
            for space_id, external_ids in space_id_to_external_ids.items()
            if len(external_ids) > 1
        }
        if duplicates:
            parts = [
                f"  space {space_id!r} <- datasets {humanize_collection(ext_ids)}"
                for space_id, ext_ids in duplicates.items()
            ]
            raise ToolkitMigrationError(
                "Multiple dataset external IDs resolve to the same space ID:\n" + "\n".join(parts)
            )

        return result

    def store_lineage(self, lineage_nodes: Sequence[NodeRequest]) -> int:
        res = self.client.tool.instances.create(lineage_nodes)
        return len(res)


class SourceSystemCreator(MigrationCreator):
    COGNITE_SOURCE_SYSTEM_VIEW_ID = ViewReference(space="cdf_cdm", external_id="CogniteSourceSystem", version="v1")

    def __init__(
        self,
        client: ToolkitClient,
        instance_space: str,
        data_set_external_id: str | None = None,
        hierarchy: str | None = None,
    ) -> None:
        super().__init__(client)
        if sum([data_set_external_id is not None, hierarchy is not None]) != 1:
            raise ValueError("Exactly one of data_set_external_id or hierarchy must be provided.")
        self._instance_space = instance_space
        self.data_set_external_id = data_set_external_id
        self.hierarchy = hierarchy

    def create_resources(self) -> Iterable[ToCreateResources]:
        existing_resources = self._get_existing_source_systems()
        seen: set[str] = set()
        to_create: list[CreatedResource[NodeRequest]] = []
        for source in self._lookup_sources():
            source_str = source.value
            if not isinstance(source_str, str) or source_str in seen:
                continue
            seen.add(source_str)
            if existing_id := existing_resources.get(source_str):
                self.client.console.print(f"Skipping {source_str} as it already exists in {existing_id!r}.")
                continue
            data_source = InstanceSource(source=self.COGNITE_SOURCE_SYSTEM_VIEW_ID, properties={"name": source_str})
            linage_source = InstanceSource(source=CREATED_SOURCE_SYSTEM_VIEW_ID, properties={"source": source_str})
            to_create.append(
                CreatedResource[NodeRequest](
                    resource=NodeRequest(
                        space=self._instance_space,
                        external_id=source_str,
                        sources=[data_source, linage_source],
                    ),
                    config_data=NodeRequest(
                        space=self._instance_space,
                        external_id=source_str,
                        sources=[data_source],
                    ).dump(),
                    filestem=source_str,
                )
            )
        yield ToCreateResources(
            resources=to_create,
            crud_cls=NodeCRUD,
            display_name="Source System",
            store_linage=lambda: len(to_create),
        )

    def _get_existing_source_systems(self) -> dict[str, NodeReference]:
        all_existing = self.client.migration.created_source_system.list(limit=-1)
        return {node.source: NodeReference(space=node.space, external_id=node.external_id) for node in all_existing}

    def _lookup_sources(self) -> Iterable[UniqueResult]:
        yield from self.client.assets.aggregate_unique_values(AssetProperty.source, filter=self._simple_filter)
        yield from self.client.events.aggregate_unique_values(property=EventProperty.source, filter=self._simple_filter)
        yield from self.client.documents.aggregate_unique_values(
            SourceFileProperty.source, filter=self._advanced_filter, limit=1000
        )

    @cached_property
    def _simple_filter(self) -> dict[str, Any] | None:
        if self.data_set_external_id is not None:
            return {"dataSetIds": [{"externalId": self.data_set_external_id}]}
        if self.hierarchy is not None:
            return {"assetSubtreeIds": [{"externalId": self.hierarchy}]}
        return None

    @cached_property
    def _advanced_filter(self) -> filters.Filter:
        if self.data_set_external_id is not None:
            data_set_id = self.client.lookup.data_sets.id(self.data_set_external_id)
            if data_set_id is None:
                raise ToolkitMissingResourceError(f"Data set with external ID '{self.data_set_external_id}' not found.")
            return filters.Equals(SourceFileProperty.data_set_id, data_set_id)
        if self.hierarchy is not None:
            return filters.InAssetSubtree(SourceFileProperty.asset_external_ids, [self.hierarchy])
        else:
            raise ValueError("This should not happen.")


class InfieldV2ConfigCreator(MigrationCreator):
    TARGET_SPACE = "APM_Config"

    def __init__(
        self,
        client: ToolkitClient,
        external_ids: Sequence[str] | None = None,
        apm_configs: Sequence[APMConfigResponse] | None = None,
    ) -> None:
        super().__init__(client)
        if sum([external_ids is not None, apm_configs is not None]) != 1:
            raise ValueError("Exactly one of external_ids or apm_configs must be provided.")
        self.external_ids = external_ids
        self.apm_configs = apm_configs

    def create_resources(self) -> Iterable[ToCreateResources]:
        if self.external_ids is not None:
            apm_configs = self.client.infield.apm_config.retrieve(
                TypedNodeIdentifier.from_str_ids(self.external_ids, space=APM_CONFIG_SPACE)
            )
        else:
            # We know this is not None from the check in __init__
            apm_configs = list(cast(Sequence[APMConfigResponse], self.apm_configs))

        all_location_configs: list[CreatedResource[InFieldCDMLocationConfigRequest]] = []
        all_location_filters: list[CreatedResource[LocationFilterRequest]] = []
        for apm_config in apm_configs:
            location_configs, location_filters = self._create_infield_v2_config(apm_config)
            all_location_configs.extend(
                CreatedResource(
                    resource=loc_config,
                    config_data=loc_config.dump(context="toolkit", exclude={"instance_type"}),
                    filestem=f"{apm_config.external_id}_location_{loc_config.name}",
                )
                for loc_config in location_configs
            )
            all_location_filters.extend(
                CreatedResource(
                    resource=loc_filter,
                    config_data=loc_filter.dump(),
                    filestem=f"{apm_config.external_id}_filter_{loc_filter.name}",
                )
                for loc_filter in location_filters
            )
        yield ToCreateResources(
            resources=all_location_filters,
            crud_cls=LocationFilterCRUD,
            display_name="Location Filters",
        )
        yield ToCreateResources(
            resources=all_location_configs,
            crud_cls=InFieldCDMLocationConfigCRUD,
            display_name="InField CDM Location Configs",
        )

    def _create_infield_v2_config(
        self, config: APMConfigResponse
    ) -> tuple[list[InFieldCDMLocationConfigRequest], list[LocationFilterRequest]]:
        location_configs: list[InFieldCDMLocationConfigRequest] = []
        location_filters: list[LocationFilterRequest] = []
        if not config.feature_configuration:
            return location_configs, location_filters

        data_exploration = self._create_data_exploration(config.feature_configuration)

        for index, root_location_config in enumerate(config.feature_configuration.root_location_configurations or []):
            location_filter = self._create_location_filter(root_location_config)
            location_filters.append(location_filter)
            location_configs.append(
                self._create_location_config(
                    root_location_config,
                    config.feature_configuration.disciplines,
                    data_exploration,
                    index,
                )
            )

        return location_configs, location_filters

    def _create_location_filter(self, config: RootLocationConfiguration) -> LocationFilterRequest:
        original_external_id = config.external_id or config.asset_external_id or str(uuid.uuid4())
        external_id = f"location_filter_{original_external_id}"
        name = config.display_name or config.asset_external_id or external_id

        instance_spaces = [
            space
            for space in [config.app_data_instance_space, config.source_data_instance_space]
            if space is not None and space != ""
        ]

        # Todo: Scene and views
        return LocationFilterRequest(
            external_id=external_id,
            name=name,
            description="InField location, migrated from old location configuration",
            instance_spaces=instance_spaces or None,
            data_models=[INFIELD_ON_CDM_DATA_MODEL],
            data_modeling_type="DATA_MODELING_ONLY",
        )

    def _create_location_config(
        self,
        config: RootLocationConfiguration,
        disciplines: list[Discipline] | None,
        data_exploration: dict[str, JsonValue],
        index: int,
    ) -> InFieldCDMLocationConfigRequest:
        if (
            config.asset_external_id is None
            or config.app_data_instance_space is None
            or config.source_data_instance_space is None
        ):
            raise ToolkitMigrationError(
                f"Root location configuration with external ID '{config.external_id}' is missing required fields for migration. "
            )
        root_node = self.client.migration.lookup.assets(external_id=config.asset_external_id)
        if not root_node:
            raise ToolkitMissingResourceError(
                f"Root asset with external ID '{config.asset_external_id}' not migrated. Cannot create location config for root location configuration with external ID '{config.external_id}'."
            )

        if config.external_id:
            external_id = config.external_id
        else:
            external_id = f"{config.asset_external_id}_{index}"

        feature_toggles: dict[str, Any] | None = None
        if isinstance(config.feature_toggles, RootLocationFeatureToggles):
            feature_toggles = config.feature_toggles.dump()
            if config.feature_toggles.observations:
                feature_toggles["observations"] = config.feature_toggles.observations.is_enabled

        access_management: dict[str, JsonValue] = {}
        if config.template_admins:
            # Todo Fetch and update groups. Jira ticket: CDF-27033
            # list[str] is a valid JsonValue
            access_management["templateAdmins"] = config.template_admins  # type: ignore[assignment]
        if config.checklist_admins:
            access_management["checklistAdmins"] = config.checklist_admins  # type: ignore[assignment]

        data_filters: dict[str, JsonValue] = {}
        for key in ["assets", "files", "timeseries", "maintenanceOrders", "operations", "notifications"]:
            # Todo Assets does not support multiple instanceSpaces.
            #    add to Validation in cdf build.
            data_filters[key] = {"instanceSpaces": [config.source_data_instance_space]}

        data_storage: dict[str, JsonValue] = {
            "rootLocation": root_node.dump(),
            "appInstanceSpace": config.app_data_instance_space,
        }
        view_mappings: dict[str, JsonValue] = {}
        for key, default_view in [
            ("asset", ViewReference(space="cdf_cdm", external_id="CogniteAsset", version="v1")),
            ("operation", ViewReference(space="cdf_idm", external_id="CogniteOperation", version="v1")),
            ("notification", ViewReference(space="cdf_idm", external_id="CogniteNotification", version="v1")),
            ("maintenanceOrder", ViewReference(space="cdf_idm", external_id="CogniteMaintenanceOrder", version="v1")),
            ("file", ViewReference(space="cdf_cdm", external_id="CogniteFile", version="v1")),
        ]:
            view_mappings[key] = default_view.dump()

        return InFieldCDMLocationConfigRequest(
            space=config.source_data_instance_space or "<Please fill in the source data instance space>",
            external_id=external_id,
            name="InField Location Config",
            description="Migrated InField Location Configuration",
            feature_toggles=feature_toggles,
            access_management=access_management or None,
            data_filters=data_filters,
            disciplines=[discipline.dump() for discipline in disciplines] if disciplines else None,
            data_storage=data_storage,
            view_mappings=view_mappings,
            data_exploration_config=data_exploration or None,
        )

    def _create_data_exploration(self, config: FeatureConfiguration) -> dict[str, JsonValue]:
        data_exploration: dict[str, JsonValue] = {}
        if config.observations:
            data_exploration["observations"] = config.observations
        if config.documents:
            documents: dict[str, JsonValue] = {}
            if config.documents.type:
                documents["type"] = config.documents.type.removeprefix("metadata.")
            if config.documents.title:
                documents["title"] = config.documents.title
            if config.documents.description:
                documents["description"] = config.documents.description.removeprefix("metadata.")
            data_exploration["documents"] = documents
        if config.notifications:
            data_exploration["notifications"] = config.notifications.dump()
        if config.asset_page_configuration:
            dumped = config.asset_page_configuration.dump()
            # Linkable Asset Key is used for implicit connections in the old Asset.metaadata.
            dumped.pop("linkableAssetKeys", None)
            data_exploration["assets"] = dumped
        return data_exploration



from abc import ABC
from pathlib import Path
from typing import Annotated, TypeAlias

from pydantic import BaseModel, ConfigDict, Field, PlainValidator
from pydantic.alias_generators import to_camel

from cognite_toolkit._cdf_tk.constants import DATA_MANIFEST_SUFFIX
from cognite_toolkit._cdf_tk.utils.file import safe_write, sanitize_filename, yaml_safe_dump
from cognite_toolkit._cdf_tk.utils.text import to_sentence_case
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal


def _check_is_sanitized_filename(value: str) -> str:
    if value != sanitize_filename(value):
        raise ValueError(f"Value {value!r} is not a sanitized filename.")
    return value


DirectoryName: TypeAlias = Annotated[str, PlainValidator(_check_is_sanitized_filename)]


class SelectorObject(BaseModel):
    """This is used as base class for all selector objects including nested ones."""

    model_config = ConfigDict(frozen=True, alias_generator=to_camel, populate_by_name=True)


class DataSelector(SelectorObject, ABC):
    """A selector gives instructions on what data to select from CDF.

    For example, for instances it can be a view or container, while for assets it can be a data set or asset subtree.

    Args:
        type: The type of selector.
        kind: The kind of data the selector is for (e.g., 'RawRows', 'Assets').
        download_dir_name: (Required for download) The name of the directory where the data files
            should be downloaded.
    """

    type: str
    kind: str
    download_dir_name: DirectoryName | None = Field(None, exclude=True)

    def dump(self) -> dict[str, JsonVal]:
        return self.model_dump(by_alias=True)

    def dump_to_file(self, directory: Path) -> Path:
        """Dumps the selector to a YAML file in the specified directory.

        The filename is derived from the string representation of the selector.

        Args:
            directory: The directory where the YAML file will be saved.
        """

        filepath = directory / f"{sanitize_filename(str(self))}{DATA_MANIFEST_SUFFIX}"
        filepath.parent.mkdir(parents=True, exist_ok=True)
        safe_write(file=filepath, content=yaml_safe_dump(self.model_dump(mode="json", by_alias=True)), encoding="utf-8")
        return filepath

    @property
    def display_name(self) -> str:
        """A human-readable name for the selector."""
        return to_sentence_case(self.kind)

    def __str__(self) -> str:
        # We want to force subclasses to implement __str__
        raise NotImplementedError()

    def find_data_files(self, input_dir: Path, manifest_file: Path) -> list[Path]:
        """Find data files in the specified input directory that match this selector.

        Args:
            input_dir: The directory to search for data files.
            manifest_file: The manifest file that describes the data files.
        Returns:
            A list of Paths to the data files that match this selector.
        """
        data_file_prefix = manifest_file.name.removesuffix(DATA_MANIFEST_SUFFIX)
        return [file for file in input_dir.glob(f"{data_file_prefix}*") if not file.name.endswith(DATA_MANIFEST_SUFFIX)]



from ._client import HTTPClient
from ._data_classes import (
    ErrorDetails,
    FailedRequest,
    FailedResponse,
    HTTPResult,
    RequestMessage,
    SuccessResponse,
)
from ._exception import ToolkitAPIError
from ._item_classes import (
    ItemsFailedRequest,
    ItemsFailedResponse,
    ItemsRequest,
    ItemsResultMessage,
    ItemsSuccessResponse,
)

__all__ = [
    "ErrorDetails",
    "FailedRequest",
    "FailedResponse",
    "HTTPClient",
    "HTTPResult",
    "ItemsFailedRequest",
    "ItemsFailedResponse",
    "ItemsRequest",
    "ItemsResultMessage",
    "ItemsSuccessResponse",
    "RequestMessage",
    "SuccessResponse",
    "ToolkitAPIError",
]



from collections.abc import Hashable, Iterable, Sequence
from pathlib import Path
from typing import Any

from cognite.client.data_classes import ClientCredentials
from cognite.client.data_classes.capabilities import Capability, HostedExtractorsAcl
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.resource_classes.hosted_extractor_destination import (
    HostedExtractorDestinationRequest,
    HostedExtractorDestinationResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.hosted_extractor_job import (
    HostedExtractorJobRequest,
    HostedExtractorJobResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.hosted_extractor_mapping import (
    HostedExtractorMappingRequest,
    HostedExtractorMappingResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.hosted_extractor_source import (
    BasicAuthenticationRequest,
    ClientCredentialAuthenticationRequest,
    EventHubSourceRequest,
    HostedExtractorSourceRequest,
    HostedExtractorSourceRequestUnion,
    HostedExtractorSourceResponseUnion,
    HTTPBasicAuthenticationRequest,
    KafkaSourceRequest,
    MQTTSourceRequest,
    RESTSourceRequest,
    ScramShaAuthenticationRequest,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotSupported
from cognite_toolkit._cdf_tk.resource_classes import (
    HostedExtractorDestinationYAML,
    HostedExtractorJobYAML,
    HostedExtractorMappingYAML,
    HostedExtractorSourceYAML,
)
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning

from .data_organization import DataSetsCRUD


class HostedExtractorSourceCRUD(
    ResourceCRUD[ExternalId, HostedExtractorSourceRequestUnion, HostedExtractorSourceResponseUnion]
):
    folder_name = "hosted_extractors"
    resource_cls = HostedExtractorSourceResponseUnion  # type: ignore[assignment]
    resource_write_cls = HostedExtractorSourceRequestUnion  # type: ignore[assignment]
    kind = "Source"
    yaml_cls = HostedExtractorSourceYAML
    _doc_base_url = "https://api-docs.cognite.com/20230101-alpha/tag/"
    _doc_url = "Sources/operation/create_sources"
    _SupportedSources = (MQTTSourceRequest, KafkaSourceRequest, RESTSourceRequest, EventHubSourceRequest)

    @property
    def display_name(self) -> str:
        return "hosted extractor sources"

    @classmethod
    def get_id(cls, item: HostedExtractorSourceRequestUnion | HostedExtractorSourceResponseUnion | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        return ExternalId(external_id=item.external_id)

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[HostedExtractorSourceRequestUnion] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [HostedExtractorsAcl.Action.Read]
            if read_only
            else [HostedExtractorsAcl.Action.Read, HostedExtractorsAcl.Action.Write]
        )

        return HostedExtractorsAcl(
            actions,
            HostedExtractorsAcl.Scope.All(),
        )

    def create(self, items: Sequence[HostedExtractorSourceRequestUnion]) -> list[HostedExtractorSourceResponseUnion]:
        return self.client.tool.hosted_extractors.sources.create(list(items))

    def retrieve(self, ids: Sequence[ExternalId]) -> list[HostedExtractorSourceResponseUnion]:
        return self.client.tool.hosted_extractors.sources.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[HostedExtractorSourceRequestUnion]) -> list[HostedExtractorSourceResponseUnion]:
        return self.client.tool.hosted_extractors.sources.update(list(items), mode="replace")

    def delete(self, ids: Sequence[ExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.hosted_extractors.sources.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[HostedExtractorSourceResponseUnion]:
        for sources in self.client.tool.hosted_extractors.sources.iterate():
            yield from sources

    def dump_resource(
        self, resource: HostedExtractorSourceResponseUnion, local: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        HighSeverityWarning(
            "Sources will always be considered different, and thus will always be redeployed."
        ).print_warning()
        return self.dump_id(self.get_id(resource))

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> HostedExtractorSourceRequestUnion:
        loaded = HostedExtractorSourceRequest.validate_python(resource)
        if not isinstance(loaded, self._SupportedSources):
            # We need to explicitly check for the supported types as we need to ensure that we know
            # what the sensitive strings are for each type of source.
            # Technically, we could support any new source type, but
            # we would risk printing sensitive strings in the terminal.
            raise ToolkitNotSupported(
                f"Hosted extractor source type {type(loaded).__name__} is not supported. "
                f"Please contact support to request support for this source type."
            )
        return loaded

    def sensitive_strings(self, item: HostedExtractorSourceRequestUnion) -> Iterable[str]:
        if isinstance(item, MQTTSourceRequest | KafkaSourceRequest | RESTSourceRequest) and item.authentication:
            yield from self._sensitive_auth_strings(item.authentication)
        if (
            isinstance(item, MQTTSourceRequest | KafkaSourceRequest)
            and item.auth_certificate
            and item.auth_certificate.key_password
        ):
            yield item.auth_certificate.key_password

    @staticmethod
    def _sensitive_auth_strings(
        auth: BasicAuthenticationRequest
        | ClientCredentialAuthenticationRequest
        | ScramShaAuthenticationRequest
        | HTTPBasicAuthenticationRequest,
    ) -> Iterable[str]:
        if isinstance(auth, BasicAuthenticationRequest | ScramShaAuthenticationRequest) and auth.password:
            yield auth.password
        elif isinstance(auth, ClientCredentialAuthenticationRequest):
            yield auth.client_secret
        elif isinstance(auth, HTTPBasicAuthenticationRequest):
            yield auth.value


class HostedExtractorDestinationCRUD(
    ResourceCRUD[ExternalId, HostedExtractorDestinationRequest, HostedExtractorDestinationResponse]
):
    folder_name = "hosted_extractors"
    resource_cls = HostedExtractorDestinationResponse
    resource_write_cls = HostedExtractorDestinationRequest
    dependencies = frozenset({DataSetsCRUD})
    kind = "Destination"
    _doc_base_url = "https://api-docs.cognite.com/20230101-alpha/tag/"
    _doc_url = "Destinations/operation/create_destinations"
    yaml_cls = HostedExtractorDestinationYAML

    def __init__(self, client: ToolkitClient, build_dir: Path | None, console: Console | None = None):
        super().__init__(client, build_dir, console)
        self._authentication_by_id: dict[str, ClientCredentials] = {}

    @property
    def display_name(self) -> str:
        return "hosted extractor destinations"

    @classmethod
    def get_id(cls, item: HostedExtractorDestinationRequest | HostedExtractorDestinationResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        return ExternalId(external_id=item.external_id)

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[HostedExtractorDestinationRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [HostedExtractorsAcl.Action.Read]
            if read_only
            else [HostedExtractorsAcl.Action.Read, HostedExtractorsAcl.Action.Write]
        )

        return HostedExtractorsAcl(
            actions,
            HostedExtractorsAcl.Scope.All(),
        )

    def create(self, items: Sequence[HostedExtractorDestinationRequest]) -> list[HostedExtractorDestinationResponse]:
        return self.client.tool.hosted_extractors.destinations.create(list(items))

    def retrieve(self, ids: Sequence[ExternalId]) -> list[HostedExtractorDestinationResponse]:
        return self.client.tool.hosted_extractors.destinations.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[HostedExtractorDestinationRequest]) -> list[HostedExtractorDestinationResponse]:
        return self.client.tool.hosted_extractors.destinations.update(list(items), mode="replace")

    def delete(self, ids: Sequence[ExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.hosted_extractors.destinations.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[HostedExtractorDestinationResponse]:
        for destinations in self.client.tool.hosted_extractors.destinations.iterate():
            yield from destinations

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> HostedExtractorDestinationRequest:
        if raw_auth := resource.pop("credentials", None):
            credentials = ClientCredentials._load(raw_auth)
            self._authentication_by_id[self.get_id(resource).external_id] = credentials
            if is_dry_run:
                resource["credentials"] = {"nonce": "dummy_nonce"}
            else:
                session = self.client.iam.sessions.create(credentials, "CLIENT_CREDENTIALS")
                resource["credentials"] = {"nonce": session.nonce}
        if ds_external_id := resource.pop("targetDataSetExternalId", None):
            resource["targetDataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        return HostedExtractorDestinationRequest.model_validate(resource)

    def dump_resource(
        self, resource: HostedExtractorDestinationResponse, local: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        HighSeverityWarning(
            "Destinations will always be considered different, and thus will always be redeployed."
        ).print_warning()
        return self.dump_id(self.get_id(resource))

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "targetDataSetId" in item:
            yield DataSetsCRUD, ExternalId(external_id=item["targetDataSetId"])

    def sensitive_strings(self, item: HostedExtractorDestinationRequest) -> Iterable[str]:
        if item.credentials:
            yield item.credentials.nonce
        id_ = self.get_id(item)
        if id_.external_id in self._authentication_by_id:
            yield self._authentication_by_id[id_.external_id].client_secret


class HostedExtractorJobCRUD(ResourceCRUD[ExternalId, HostedExtractorJobRequest, HostedExtractorJobResponse]):
    folder_name = "hosted_extractors"
    resource_cls = HostedExtractorJobResponse
    resource_write_cls = HostedExtractorJobRequest
    dependencies = frozenset({HostedExtractorSourceCRUD, HostedExtractorDestinationCRUD})
    kind = "Job"
    yaml_cls = HostedExtractorJobYAML
    _doc_base_url = "https://api-docs.cognite.com/20230101-alpha/tag/"
    _doc_url = "Jobs/operation/create_jobs"

    @property
    def display_name(self) -> str:
        return "hosted extractor jobs"

    @classmethod
    def get_id(cls, item: HostedExtractorJobRequest | HostedExtractorJobResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        return ExternalId(external_id=item.external_id)

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[HostedExtractorJobRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [HostedExtractorsAcl.Action.Read]
            if read_only
            else [HostedExtractorsAcl.Action.Read, HostedExtractorsAcl.Action.Write]
        )

        return HostedExtractorsAcl(
            actions,
            HostedExtractorsAcl.Scope.All(),
        )

    def dump_resource(
        self, resource: HostedExtractorJobResponse, local: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        local = local or {}
        if not dumped.get("config") and "config" not in local:
            dumped.pop("config", None)
        return dumped

    def create(self, items: Sequence[HostedExtractorJobRequest]) -> list[HostedExtractorJobResponse]:
        return self.client.tool.hosted_extractors.jobs.create(list(items))

    def retrieve(self, ids: Sequence[ExternalId]) -> list[HostedExtractorJobResponse]:
        return self.client.tool.hosted_extractors.jobs.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[HostedExtractorJobRequest]) -> list[HostedExtractorJobResponse]:
        return self.client.tool.hosted_extractors.jobs.update(list(items), mode="replace")

    def delete(self, ids: Sequence[ExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.hosted_extractors.jobs.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[HostedExtractorJobResponse]:
        for jobs in self.client.tool.hosted_extractors.jobs.iterate():
            yield from jobs

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "sourceId" in item:
            yield HostedExtractorSourceCRUD, ExternalId(external_id=item["sourceId"])
        if "destinationId" in item:
            yield HostedExtractorDestinationCRUD, ExternalId(external_id=item["destinationId"])


class HostedExtractorMappingCRUD(
    ResourceCRUD[ExternalId, HostedExtractorMappingRequest, HostedExtractorMappingResponse]
):
    folder_name = "hosted_extractors"
    resource_cls = HostedExtractorMappingResponse
    resource_write_cls = HostedExtractorMappingRequest
    # This is not an explicit dependency, however, adding it here as mapping will should be deployed after source.
    dependencies = frozenset({HostedExtractorSourceCRUD})
    kind = "Mapping"
    _doc_base_url = "https://api-docs.cognite.com/20230101-alpha/tag/"
    _doc_url = "Mappings/operation/create_mappings"
    yaml_cls = HostedExtractorMappingYAML

    @property
    def display_name(self) -> str:
        return "hosted extractor mappings"

    @classmethod
    def get_id(cls, item: HostedExtractorMappingRequest | HostedExtractorMappingResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        return ExternalId(external_id=item.external_id)

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[HostedExtractorMappingRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [HostedExtractorsAcl.Action.Read]
            if read_only
            else [HostedExtractorsAcl.Action.Read, HostedExtractorsAcl.Action.Write]
        )

        return HostedExtractorsAcl(
            actions,
            HostedExtractorsAcl.Scope.All(),
        )

    def create(self, items: Sequence[HostedExtractorMappingRequest]) -> list[HostedExtractorMappingResponse]:
        return self.client.tool.hosted_extractors.mappings.create(list(items))

    def retrieve(self, ids: Sequence[ExternalId]) -> list[HostedExtractorMappingResponse]:
        return self.client.tool.hosted_extractors.mappings.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[HostedExtractorMappingRequest]) -> list[HostedExtractorMappingResponse]:
        return self.client.tool.hosted_extractors.mappings.update(list(items), mode="replace")

    def delete(self, ids: Sequence[ExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.hosted_extractors.mappings.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[HostedExtractorMappingResponse]:
        for mappings in self.client.tool.hosted_extractors.mappings.iterate():
            yield from mappings



"""Base classes for resource clients.

This module provides the base infrastructure for resource-specific clients
that handle CRUD operations for CDF Data Modeling API resources.
"""

from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Callable, Iterable, Sequence
from dataclasses import dataclass
from functools import partial
from typing import Any, Generic, Literal, TypeAlias

from pydantic import JsonValue

from cognite_toolkit._cdf_tk.client._resource_base import (
    RequestItem,
    T_BaseModelObject,
    UpdatableRequestResource,
)
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsRequest,
    ItemsSuccessResponse,
    RequestMessage,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence

from .responses import PagedResponse


@dataclass(frozen=True)
class Endpoint:
    method: Literal["GET", "POST", "PUT", "PATCH", "DELETE"]
    path: str
    item_limit: int
    concurrency_max_workers: int = 1


APIMethod: TypeAlias = Literal["create", "retrieve", "update", "delete", "list", "upsert"]


class CDFResourceAPI(Generic[T_BaseModelObject], ABC):
    """Generic resource API for CDF APIs

    This class provides the logic for working with CDF resources,
    including creating, retrieving, deleting, and listing resources.
    """

    def __init__(
        self,
        http_client: HTTPClient,
        method_endpoint_map: dict[APIMethod, Endpoint],
        disable_gzip: bool = False,
        api_version: str | None = None,
    ) -> None:
        """Initialize the resource API.

        Args:
            http_client: The HTTP client to use for API requests.
            method_endpoint_map: A mapping of endpoint suffixes to their properties.
            disable_gzip: Whether to disable gzip compression for requests. Defaults to False.
                This is only used by the robotics API. If that API is dropped, this parameter can be removed.
        """
        self._http_client = http_client
        self._method_endpoint_map = method_endpoint_map
        self._disable_gzip = disable_gzip
        self._api_version = api_version

    @classmethod
    def _serialize_items(cls, items: Sequence[RequestItem]) -> list[dict[str, JsonValue]]:
        """Serialize reference objects to JSON-compatible dicts."""
        return [item.dump() for item in items]

    @classmethod
    def _serialize_updates(
        cls, items: Sequence[UpdatableRequestResource], mode: Literal["patch", "replace"]
    ) -> list[dict[str, JsonValue]]:
        """Serialize updateable objects to JSON-compatible dicts."""
        return [item.as_update(mode=mode) for item in items]

    @abstractmethod
    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[T_BaseModelObject]:
        """Parse a single item response."""
        raise NotImplementedError()

    def _make_url(self, path: str = "") -> str:
        """Create the full URL for this resource endpoint."""
        return self._http_client.config.create_api_url(path)

    def _update(
        self, items: Sequence[UpdatableRequestResource], mode: Literal["patch", "replace"]
    ) -> list[T_BaseModelObject]:
        """Update resources in chunks."""
        response_items: list[T_BaseModelObject] = []
        for response in self._chunk_requests(
            items, "update", serialization=partial(self._serialize_updates, mode=mode)
        ):
            response_items.extend(self._validate_page_response(response).items)
        return response_items

    def _request_item_response(
        self,
        items: Sequence[RequestItem],
        method: APIMethod,
        params: dict[str, Any] | None = None,
        extra_body: dict[str, Any] | None = None,
        endpoint: str | None = None,
    ) -> list[T_BaseModelObject]:
        response_items: list[T_BaseModelObject] = []
        for response in self._chunk_requests(items, method, self._serialize_items, params, extra_body, endpoint):
            response_items.extend(self._validate_page_response(response).items)
        return response_items

    def _request_no_response(
        self,
        items: Sequence[RequestItem],
        method: APIMethod,
        params: dict[str, Any] | None = None,
        extra_body: dict[str, Any] | None = None,
        endpoint: str | None = None,
    ) -> None:
        list(self._chunk_requests(items, method, self._serialize_items, params, extra_body, endpoint))
        return None

    def _chunk_requests(
        self,
        items: Sequence[RequestItem],
        method: APIMethod,
        serialization: Callable[[Sequence[RequestItem]], list[dict[str, JsonValue]]],
        params: dict[str, Any] | None = None,
        extra_body: dict[str, Any] | None = None,
        endpoint_path: str | None = None,
    ) -> Iterable[SuccessResponse]:
        """Send requests in chunks and yield responses.

        Args:
            items: The items to process.
            method: The API method to use. This is used ot look the up the endpoint.
            serialization: A function to serialize the items to JSON-compatible dicts.
            params: Optional query parameters for the request.
            extra_body: Optional extra body content to include in the request.
            endpoint_path: Optional override for the endpoint path.

        Yields:
            The successful responses from the API.
        """
        # Filter out None
        request_params = self._filter_out_none_values(params)
        endpoint = self._method_endpoint_map[method]

        for chunk in chunker_sequence(items, endpoint.item_limit):
            request = RequestMessage(
                endpoint_url=f"{self._make_url(endpoint_path or endpoint.path)}",
                method=endpoint.method,
                body_content={"items": serialization(chunk), **(extra_body or {})},  # type: ignore[dict-item]
                parameters=request_params,
                disable_gzip=self._disable_gzip,
                api_version=self._api_version,
            )
            response = self._http_client.request_single_retries(request)
            yield response.get_success_or_raise()

    def _request_item_split_retries(
        self,
        items: Sequence[RequestItem],
        method: APIMethod,
        params: dict[str, Any] | None = None,
        extra_body: dict[str, Any] | None = None,
    ) -> list[T_BaseModelObject]:
        """Request items with retries, splitting on failures.

        This method handles large batches of items by chunking them according to the endpoint's item limit.
        If a single item fails, it splits the request into individual item requests to isolate the failure.

        Args:
            items: Sequence of items to request.
            method: API method to use for the request.
            params: Optional query parameters for the request.
            extra_body: Optional additional body fields for the request.
        Returns:
            List of response items.
        """
        response_items: list[T_BaseModelObject] = []
        for response in self._chunk_requests_items_split_retries(items, method, params, extra_body):
            response_items.extend(self._validate_page_response(response).items)
        return response_items

    def _request_item_split_retries_no_response(
        self,
        items: Sequence[RequestItem],
        method: APIMethod,
        params: dict[str, Any] | None = None,
        extra_body: dict[str, Any] | None = None,
    ) -> None:
        """Request items with retries, splitting on failures, without returning any response.

        This method handles large batches of items by chunking them according to the endpoint's item limit.
        If a single item fails, it splits the request into individual item requests to isolate the failure.

        Args:
            items: Sequence of items to request.
            method: API method to use for the request.
            params: Optional query parameters for the request.
            extra_body: Optional additional body fields for the request.
        """
        list(self._chunk_requests_items_split_retries(items, method, params, extra_body))
        return None

    def _chunk_requests_items_split_retries(
        self,
        items: Sequence[RequestItem],
        method: APIMethod,
        params: dict[str, Any] | None = None,
        extra_body: dict[str, Any] | None = None,
    ) -> Iterable[ItemsSuccessResponse]:
        """Request items with retries and splitting on failures.

        This method handles large batches of items by chunking them according to the endpoint's item limit.
        If a single item fails, it splits the request into individual item requests to isolate the failure.

        This can be useful to create ignore unknown IDs behavior when retrieving items for API endpoints that
        do not support it natively.

        Args:
            items: Sequence of items to request.
            method: API method to use for the request.
            params: Optional query parameters for the request.
            extra_body: Optional additional body fields for the request.
        Yields
            SuccessResponse2: Successful responses from the API. All failed items are skipped.

        """
        # Filter out None
        request_params = self._filter_out_none_values(params)
        endpoint = self._method_endpoint_map[method]

        for chunk in chunker_sequence(items, endpoint.item_limit):
            request = ItemsRequest(
                endpoint_url=f"{self._make_url(endpoint.path)}",
                method=endpoint.method,
                parameters=request_params,
                items=chunk,
                extra_body_fields=extra_body,
                disable_gzip=self._disable_gzip,
                api_version=self._api_version,
            )
            responses = self._http_client.request_items_retries(request)
            for response in responses:
                if isinstance(response, ItemsSuccessResponse):
                    yield response

    @classmethod
    def _filter_out_none_values(cls, params: dict[str, Any] | None) -> dict[str, Any] | None:
        request_params: dict[str, Any] | None = None
        if params:
            request_params = {k: v for k, v in params.items() if v is not None}
        return request_params

    @classmethod
    def _group_items_by_text_field(
        cls, items: Sequence[RequestItem], *field_names: str
    ) -> dict[tuple[str, ...], list[RequestItem]]:
        """Group items by a text field."""
        grouped_items: dict[tuple[str, ...], list[RequestItem]] = defaultdict(list)
        for item in items:
            key = tuple(str(getattr(item, field_name)) for field_name in field_names)
            grouped_items[key].append(item)
        return grouped_items

    def _paginate(
        self,
        limit: int,
        cursor: str | None = None,
        params: dict[str, Any] | None = None,
        body: dict[str, Any] | None = None,
        endpoint_path: str | None = None,
    ) -> PagedResponse[T_BaseModelObject]:
        """Fetch a single page of resources.

        Args:
            params: Query parameters for the request. Supported parameters depend on
                the resource type but typically include:
                - cursor: Cursor for pagination
                - limit: Maximum number of items (defaults to list limit)
                - space: Filter by space
                - includeGlobal: Whether to include global resources
            body : Body content for the request, if applicable.
                limit: Maximum number of items to return in the page.
                cursor: Cursor for pagination.
            limit: Maximum number of items to return in the page.
            cursor: Cursor for pagination.

        Returns:
            A Page containing the items and the cursor for the next page.
        """
        endpoint = self._method_endpoint_map["list"]
        if not (0 < limit <= endpoint.item_limit):
            raise ValueError(f"Limit must be between 1 and {endpoint.item_limit}, got {limit}.")

        request_params = self._filter_out_none_values(params) or {}
        body = self._filter_out_none_values(body) or {}
        if endpoint.method == "GET":
            request_params["limit"] = limit
            if cursor is not None:
                request_params["cursor"] = cursor
        elif endpoint.method in {"POST", "PUT", "PATCH"}:
            body["limit"] = limit
            if cursor is not None:
                body["cursor"] = cursor
        else:
            raise NotImplementedError(f"Unsupported method {endpoint.method} for pagination.")

        request = RequestMessage(
            endpoint_url=self._make_url(endpoint_path or endpoint.path),
            method=endpoint.method,
            parameters=request_params,
            body_content=body,
            disable_gzip=self._disable_gzip,
            api_version=self._api_version,
        )
        result = self._http_client.request_single_retries(request)
        response = result.get_success_or_raise()
        return self._validate_page_response(response)

    def _iterate(
        self,
        limit: int | None = None,
        cursor: str | None = None,
        params: dict[str, Any] | None = None,
        body: dict[str, Any] | None = None,
        endpoint_path: str | None = None,
    ) -> Iterable[list[T_BaseModelObject]]:
        """Iterate over all resources, handling pagination automatically."""
        next_cursor = cursor
        total = 0
        endpoint = self._method_endpoint_map["list"]
        while True:
            page_limit = endpoint.item_limit if limit is None else min(limit - total, endpoint.item_limit)
            page = self._paginate(
                limit=page_limit, cursor=next_cursor, params=params, body=body, endpoint_path=endpoint_path
            )
            yield page.items
            total += len(page.items)
            if page.next_cursor is None or (limit is not None and total >= limit) or not page.items:
                break
            next_cursor = page.next_cursor

    def _list(
        self,
        limit: int | None = None,
        params: dict[str, Any] | None = None,
        endpoint_path: str | None = None,
        body: dict[str, Any] | None = None,
    ) -> list[T_BaseModelObject]:
        """List all resources, handling pagination automatically."""
        return [
            item
            for batch in self._iterate(limit=limit, params=params, endpoint_path=endpoint_path, body=body)
            for item in batch
        ]



from collections.abc import Iterable, Sequence
from typing import Any, Literal

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.dataset import DataSetRequest, DataSetResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalOrExternalId


class DataSetsAPI(CDFResourceAPI[DataSetResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/datasets", item_limit=10, concurrency_max_workers=1),
                "retrieve": Endpoint(method="POST", path="/datasets/byids", item_limit=1000, concurrency_max_workers=1),
                "update": Endpoint(method="POST", path="/datasets/update", item_limit=1000, concurrency_max_workers=1),
                "list": Endpoint(method="POST", path="/datasets/list", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[DataSetResponse]:
        return PagedResponse[DataSetResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[InternalOrExternalId]:
        return ResponseItems[InternalOrExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[DataSetRequest]) -> list[DataSetResponse]:
        """Create data sets in CDF.

        Args:
            items: List of DataSetRequest objects to create.
        Returns:
            List of created DataSetResponse objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(
        self, items: Sequence[InternalOrExternalId], ignore_unknown_ids: bool = False
    ) -> list[DataSetResponse]:
        """Retrieve data sets from CDF.

        Args:
            items: List of InternalOrExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        Returns:
            List of retrieved DataSetResponse objects.
        """
        return self._request_item_response(
            items, method="retrieve", extra_body={"ignoreUnknownIds": ignore_unknown_ids}
        )

    def update(
        self, items: Sequence[DataSetRequest], mode: Literal["patch", "replace"] = "replace"
    ) -> list[DataSetResponse]:
        """Update data sets in CDF.

        Args:
            items: List of DataSetRequest objects to update.
            mode: Update mode, either "patch" or "replace".

        Returns:
            List of updated DataSetResponse objects.
        """
        return self._update(items, mode=mode)

    def paginate(
        self,
        metadata: dict[str, str] | None = None,
        external_id_prefix: str | None = None,
        write_protected: bool | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[DataSetResponse]:
        """Iterate over all data sets in CDF.

        Args:
            metadata: Filter by metadata.
            external_id_prefix: Filter by external ID prefix.
            write_protected: Filter by write protection status.
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of DataSetResponse objects.
        """
        filter_body: dict[str, Any] = {}
        if metadata is not None:
            filter_body["metadata"] = metadata
        if external_id_prefix is not None:
            filter_body["externalIdPrefix"] = external_id_prefix
        if write_protected is not None:
            filter_body["writeProtected"] = write_protected

        return self._paginate(
            cursor=cursor,
            limit=limit,
            body={"filter": filter_body} if filter_body else {},
        )

    def iterate(
        self,
        metadata: dict[str, str] | None = None,
        external_id_prefix: str | None = None,
        write_protected: bool | None = None,
        limit: int | None = 100,
    ) -> Iterable[list[DataSetResponse]]:
        """Iterate over all data sets in CDF.

        Args:
            metadata: Filter by metadata.
            external_id_prefix: Filter by external ID prefix.
            write_protected: Filter by write protection status.
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of DataSetResponse objects.
        """
        filter_body: dict[str, Any] = {}
        if metadata is not None:
            filter_body["metadata"] = metadata
        if external_id_prefix is not None:
            filter_body["externalIdPrefix"] = external_id_prefix
        if write_protected is not None:
            filter_body["writeProtected"] = write_protected

        return self._iterate(
            limit=limit,
            body={"filter": filter_body} if filter_body else {},
        )

    def list(self, limit: int | None = 100) -> list[DataSetResponse]:
        """List all data sets in CDF.

        Returns:
            List of DataSetResponse objects.
        """
        return self._list(limit=limit)
